import os
import anndata
from SpaDOT.utils import _analyze_utils

def analyze(args):
    data_dir = os.path.abspath(args.data)
    if not args.output_dir:
        args.output_dir = os.path.dirname(data_dir)

    latent = anndata.read_h5ad(data_dir)
    if args.n_clusters is None:
        latent = _analyze_utils.Adaptive_clustering(args, latent)
    else:
        latent = _analyze_utils.KMeans_Clustering(latent, args.n_clusters)
    latent.obs['pixel_x'] = latent.obsm['spatial'][:, 0]
    latent.obs['pixel_y'] = latent.obsm['spatial'][:, 1]
    # draw domains
    _analyze_utils.plot_domains(args, latent)
    # perform optimal transport analysis
    _analyze_utils.OT_analysis(args, latent)
    # plot OT results
    _analyze_utils.plot_OT(args, latent)


if __name__ == "__main__":
    data_dir = "./examples"
    # create arguments for testing
    class Args:
        data = os.path.join(data_dir, "latent.h5ad")
        prefix = ""
        # n_clusters = [5, 7, 7, 6]
        n_clusters = None
    args = Args()
    # create output directory if not exists
    if 'output_dir' not in args.__dict__:
        args.output_dir = os.path.dirname(args.data)
    if not os.path.exists(args.output_dir):
        os.makedirs(args.output_dir)
    # analyze latent representations
    analyze(args)


import torch
import torch.nn as nn
import torch.nn.functional as F
from .encoder import SVGPEncoder, GATEncoder
from .decoder import Decoder
from .svgp import SVGP

class SpaDOT(nn.Module):
    def __init__(self, model_config, dataloader_dict):
        super(SpaDOT, self).__init__()
        self.input_dim = model_config['input_dim']
        self.SVGP_z_dim = model_config['z_dim'] // 2
        self.GAT_z_dim = model_config['z_dim'] // 2
        self.dtype = model_config['dtype']
        self.device = model_config['device']

        # Build encoders
        self.SVGPEncoder = SVGPEncoder(
            input_dim=self.input_dim,
            SVGP_z_dim=self.SVGP_z_dim,
            hidden_dims=model_config['svgp_encoder_layers']
        ).to(dtype=self.dtype)

        self.GATEncoder = GATEncoder(
            input_dim=self.input_dim,
            GAT_z_dim=self.GAT_z_dim,
            hidden_dim=model_config['gat_encoder_hidden'],
            num_heads=model_config['gat_attention_heads']
        ).to(dtype=self.dtype)

        # Build decoder
        self.decoder = Decoder(
            input_dim=self.input_dim,
            z_dim=self.SVGP_z_dim + self.GAT_z_dim, # in case input z_dim is odd
            decoder_layers=model_config['decoder_layers']
        ).to(dtype=self.dtype)

        # Build SVGP
        self.svgp_dict = nn.ModuleDict({
            str(tp): SVGP(
                model_config=model_config,
                inducing_points=dataloader_dict['inducing_points'][tp],
                N_train=dataloader_dict['N_train'][tp]
            ).to(model_config['device']) for tp in model_config['timepoints']
        })
        # Kmeans/OT-related parameters
        self.gammas = {}
        self.kmeans_center_dict = {}
        self.kmeans_cluster_dict = {}
        self.kmeans_index_dict = {}

    def forward(self, x, y, edge_index, tp, batch_size):
        # SVGP latent
        SVGP_qnet_mu, SVGP_qnet_var = self.SVGPEncoder(y[:batch_size])
        inside_elbo_recon, inside_elbo_kl = [], []
        SVGP_p_m, SVGP_p_v = [], []
        for l in range(self.SVGP_z_dim):
            p_m_l, p_v_l, mu_hat_l, A_hat_l = self.svgp_dict[str(tp)].approximate_posterior_params(x[:batch_size], x[:batch_size],
                                                                    SVGP_qnet_mu[:, l], SVGP_qnet_var[:, l])
            inside_elbo_recon_l, inside_elbo_kl_l = self.svgp_dict[str(tp)].variational_loss(x=x[:batch_size], y=SVGP_qnet_mu[:, l],
                                                                    noise=SVGP_qnet_var[:, l], mu_hat=mu_hat_l,
                                                                    A_hat=A_hat_l)
            inside_elbo_recon.append(inside_elbo_recon_l)
            inside_elbo_kl.append(inside_elbo_kl_l)
            SVGP_p_m.append(p_m_l)
            SVGP_p_v.append(p_v_l)
        inside_elbo_recon = torch.stack(inside_elbo_recon, dim=-1)
        inside_elbo_kl = torch.stack(inside_elbo_kl, dim=-1)
        inside_elbo_recon = torch.sum(inside_elbo_recon)
        inside_elbo_kl = torch.sum(inside_elbo_kl)
        inside_elbo = inside_elbo_recon - (batch_size / self.svgp_dict[str(tp)].N_train) * inside_elbo_kl
        SVGP_p_m = torch.stack(SVGP_p_m, dim=1)
        SVGP_p_v = torch.stack(SVGP_p_v, dim=1)
        ce_term = self._gauss_cross_entropy(SVGP_p_m, SVGP_p_v, SVGP_qnet_mu, SVGP_qnet_var)
        ce_term = torch.sum(ce_term)
        diff = ce_term - inside_elbo
        SVGP_KL = (-diff if ce_term.item() > inside_elbo.item() else diff) / self.SVGP_z_dim #  force KL to be negative (Recon - beta*KL), stablize training
        SVGP_latent_sample = SVGP_p_m + torch.randn_like(SVGP_p_m) * torch.sqrt(SVGP_p_v)

        # GAT latent
        GAT_m, GAT_v = self.GATEncoder(y, edge_index)
        GAT_m, GAT_v = GAT_m[:batch_size, :], GAT_v[:batch_size, :]
        GAT_latent_sample = GAT_m + torch.randn_like(GAT_m) * torch.sqrt(GAT_v)
        GAT_KL = -0.5 * torch.sum(1 + torch.log(GAT_v) - GAT_m.pow(2) - GAT_v) / self.GAT_z_dim

        # Concatenated latent samples
        final_latent = torch.cat([SVGP_latent_sample, GAT_latent_sample], dim=1)
        # Reconstruction loss
        recon_loss = torch.sum((y[:batch_size]-self.decoder(final_latent))**2) / self.input_dim
        # Alignment loss
        alignment_loss = F.mse_loss(SVGP_latent_sample.norm(dim=1) / self.SVGP_z_dim, 
                                    GAT_latent_sample.norm(dim=1) / self.GAT_z_dim, 
                                    reduction='sum')
        return recon_loss, SVGP_KL, GAT_KL, alignment_loss, final_latent
    
    def all_latent_samples(self, X, Y, edge_index, tp):
        """
        Output latent embedding.

        Parameters:
        -----------
        X: array_like, shape (n_spots, 2)
            Location information.
        Y: array_like, shape (n_spots, n_genes)
            Preprocessed count matrix.
        """ 
        # solve warinings
        X = torch.tensor(X, dtype=self.dtype, device=self.device)
        Y = torch.tensor(Y, dtype=self.dtype, device=self.device)
        edge_index = torch.tensor(edge_index, dtype=torch.long, device=self.device)
        SVGP_qnet_mu, SVGP_qnet_var = self.SVGPEncoder(Y)
        SVGP_p_m, SVGP_p_v = [], []
        for l in range(self.SVGP_z_dim):
            p_m_l, p_v_l, _, _ = self.svgp_dict[str(tp)].approximate_posterior_params(X, X, SVGP_qnet_mu[:, l], SVGP_qnet_var[:, l])
            SVGP_p_m.append(p_m_l)
            SVGP_p_v.append(p_v_l)
        SVGP_p_m = torch.stack(SVGP_p_m, dim=1)
        SVGP_p_v = torch.stack(SVGP_p_v, dim=1)

        GAT_m, _ = self.GATEncoder(Y, edge_index)
        p_m = torch.cat((SVGP_p_m, GAT_m), dim=1)
        latent_samples = p_m.data.cpu().detach().numpy()
        return latent_samples
    
    def _gauss_cross_entropy(self, mu1, var1, mu2, var2):
        """
        Computes the element-wise cross entropy
        Given q(z) ~ N(z| mu1, var1)
        returns E_q[ log N(z| mu2, var2) ]
        args:
            mu1:  mean of expectation (batch, tmax, 2) tf variable
            var1: var  of expectation (batch, tmax, 2) tf variable
            mu2:  mean of integrand (batch, tmax, 2) tf variable
            var2: var of integrand (batch, tmax, 2) tf variable
        returns:
            cross_entropy: (batch, tmax, 2) tf variable
        """
        term0 = 1.8378770664093453  # log(2*pi)
        term1 = torch.log(var2)
        term2 = (var1 + mu1 ** 2 - 2 * mu1 * mu2 + mu2 ** 2) / var2
        cross_entropy = -0.5 * (term0 + term1 + term2)
        return cross_entropy



import os
import anndata
import numpy as np
import pandas as pd
from sklearn.cluster import KMeans
import wot # Waddington OT, original package
import matplotlib.pyplot as plt
import seaborn as sns

def KMeans_Clustering(adata, n_clusters):
    """
    Perform KMeans clustering on the spatial data.
    
    Parameters
    ----------
    adata : anndata.AnnData
        The AnnData object containing the spatial data.
    n_clusters : int
        The number of clusters to form.
    
    Returns
    -------
    adata : anndata.AnnData
        The AnnData object with the clustering results added to the obs.
    """
    tps = adata.obs['timepoint'].unique()
    tps.sort()
    tp_adata_list = []
    for i, tp in enumerate(tps):
        # perform KMeans clustering
        tp_adata = adata[adata.obs['timepoint'] == tp].copy()
        tp_kmeans = KMeans(n_clusters=n_clusters[i], 
                           random_state=1993, 
                           n_init=10).fit(tp_adata.X)
        tp_adata.obs['kmeans'] = tp_kmeans.labels_.astype(str)
        tp_adata_list.append(tp_adata)
    # merge all timepoints
    merged_adata = anndata.concat(tp_adata_list, axis=0)
    return merged_adata


def Adaptive_clustering(args, adata, min_clusters=4, max_clusters=20, wss_threshold=0.1):
    """
    Perform adaptive clustering using KMeans and calculate WSS (Within-Cluster Sum of Squares).
    Then calculate the difference between consecutive clusters and the ratio between consecutive differences.

    Parameters:
    -----------
    adata : anndata.AnnData
        The AnnData object containing the data to cluster.
    max_clusters : int, optional
        The maximum number of clusters to test. Default is 20.
    min_clusters : int, optional
        The minimum number of clusters to test. Default is 5.

    Returns:
    --------
    dict
        A dictionary containing WSS values, differences, and ratios for each number of clusters.
    """

    tps = adata.obs['timepoint'].unique()
    tps.sort()
    tp_adata_list = []
    for i, tp in enumerate(tps):
        # perform KMeans clustering
        tp_adata = adata[adata.obs['timepoint'] == tp].copy()
        # Store WSS for each number of clusters
        wss = []
        # Perform KMeans clustering for each number of clusters
        for k in range(min_clusters, max_clusters + 1):
            kmeans = KMeans(n_clusters=k, random_state=1993, n_init=10).fit(tp_adata.X)
            wss.append(kmeans.inertia_)
        # Calculate wss and ratios
        wss_diff = -np.diff(wss)
        wss_diff_ratios = [wss_diff[i] / wss_diff[i + 1] for i in range(len(wss_diff)-1)]
        wss_df = pd.DataFrame({
            'clusters': range(min_clusters, max_clusters + 1),
            'wss': wss,
            'wss_diff': [None] + list(wss_diff),
            'wss_diff_ratio': [None] + list(wss_diff_ratios) + [None]
        })
        wss_range = wss_df['wss'].max() - wss_df['wss'].min()
        wss_diff_threshold = wss_threshold * wss_range
        filtered_wss_df = wss_df[wss_df['wss_diff'] > wss_diff_threshold]
        max_idx = filtered_wss_df['wss_diff_ratio'].idxmax()
        wss_cluster = filtered_wss_df['clusters'][max_idx]
        highlight_wss = filtered_wss_df['wss'][max_idx]
        # plot wss per cluster
        plt.figure(figsize=(10, 6))
        plt.plot(wss_df['clusters'], wss_df['wss'], marker='o')
        plt.scatter(wss_cluster, highlight_wss, color='red', s=100, label='Selected Cluster')
        plt.title('WSS vs Number of Clusters')
        plt.xlabel('Number of Clusters')
        plt.ylabel('WSS')
        plt.xticks(wss_df['clusters'])
        plt.grid()
        plt.savefig(args.output_dir+os.sep+args.prefix+str(tp)+'_WSS_vs_Clusters.png')
        plt.close()
        tp_adata.obs['kmeans'] = KMeans(n_clusters=wss_cluster, 
                                        random_state=1993, n_init=10).fit(tp_adata.X).labels_.astype(str)
        tp_adata_list.append(tp_adata)
    # merge all timepoints
    merged_adata = anndata.concat(tp_adata_list, axis=0)
    return merged_adata


def OT_analysis(args, adata):
    """
    Perform optimal transport analysis on the spatial data.
    
    Parameters
    ----------
    adata : anndata.AnnData
        The AnnData object containing the spatial data.    
    Returns
    -------
    adata : anndata.AnnData
        The AnnData object with the OT analysis results added to the obs.
    """
    adata.obs['day'] = adata.obs['timepoint'].astype('category').cat.codes
    adata.obs['cell_growth_rate'] = 1 # initialize as 1
    ot_model = wot.ot.OTModel(adata, epsilon = 0.01, lambda1 = 0.1,lambda2 = 5, growth_iters=3) # imbalanced OT with imbalanced row and comparatiely balanced target
    ot_model.compute_all_transport_maps(tmap_out=args.output_dir+os.sep+'OT') # compute transport maps
    tmap_model = wot.tmap.TransportMapModel.from_directory(args.output_dir+os.sep+'OT')
    # generate region dict
    adata.obs['SpaDOT_pred_labels'] = adata.obs['timepoint'].astype('str')+'_'+adata.obs['kmeans'].astype('str')
    latent_cell_sets = adata.obs.groupby('SpaDOT_pred_labels').apply(lambda x: x.index.tolist()).to_dict()
    days = adata.obs['day'].unique()
    days.sort()
    for tp_i in range(len(days)-1):
        prev_day = days[tp_i]
        next_day = days[tp_i+1]
        prev_day_populations = tmap_model.population_from_cell_sets(latent_cell_sets, at_time=prev_day)
        next_day_populations = tmap_model.population_from_cell_sets(latent_cell_sets, at_time=next_day)
        transition_table = tmap_model.transition_table(prev_day_populations, next_day_populations) # aggregated OT matrix
        transition_table.write_h5ad(args.output_dir+os.sep+'transition_table_'+str(prev_day)+'_'+str(next_day)+'.h5ad')

def plot_domains(args, adata):
    """
    Plot the spatial domains of the cells.
    
    Parameters
    ----------
    adata : anndata.AnnData
        The AnnData object containing the spatial data.    
    Returns
    -------
    None
    """
    for tp in adata.obs['timepoint'].unique():
        tp_adata = adata[adata.obs['timepoint'] == tp].copy()
        tp_adata.obs['pixel_x'] = tp_adata.obsm['spatial'][:, 0]
        tp_adata.obs['pixel_y'] = tp_adata.obsm['spatial'][:, 1]
        # plot the spatial domains
        plt.figure(figsize=(5, 5))
        sns.scatterplot(data=tp_adata.obs, x='pixel_x', y='pixel_y', 
                        hue='kmeans', palette='tab10', s=10)
        plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
        plt.title('Time point: {}'.format(tp))
        plt.tight_layout()
        plt.savefig(args.output_dir+os.sep+args.prefix+str(tp)+'_domains.png')
        plt.close()

def plot_OT(args, adata):
    """
    Plot the optimal transport results.
    
    Parameters
    ----------
    adata : anndata.AnnData
        The AnnData object containing the spatial data.    
    Returns
    -------
    None
    """
    days = adata.obs['day'].unique()
    days.sort()
    for tp_i in range(len(days)-1):
        prev_day = days[tp_i]
        next_day = days[tp_i+1]
        transition_table = anndata.read_h5ad(args.output_dir+os.sep+'transition_table_'+str(prev_day)+'_'+str(next_day)+'.h5ad')
        # --- normalize by column sum
        transition_prob = transition_table.X
        transition_prob = transition_prob/transition_prob.sum(axis=0, keepdims=True)
        transition_prob_col_norm_df = pd.DataFrame(transition_prob, index=transition_table.obs_names, columns=transition_table.var_names)
        # --- normalize by row sum
        transition_prob = transition_table.X
        transition_prob = transition_prob/transition_prob.sum(axis=1, keepdims=True)
        transition_prob_row_norm_df = pd.DataFrame(transition_prob, index=transition_table.obs_names, columns=transition_table.var_names)
        # --- obtain minimum of both normalized transition probabilities
        transition_prob_min_df = np.minimum(transition_prob_col_norm_df.values, transition_prob_row_norm_df.values)
        # Generate a dotplot for transition_prob_min_df with dynamic figure size
        plt.figure(figsize=(transition_prob_min_df.shape[1]*0.8, transition_prob_min_df.shape[0]*0.8))
        for i in range(transition_prob_min_df.shape[0]):
            for j in range(transition_prob_min_df.shape[1]):
                value = transition_prob_min_df[i, j]
                color = 'grey' if value < 0.2 else plt.cm.Reds(value)
                plt.scatter(j, i, s=value * 500, c=[color], edgecolors='black', alpha=0.8)
        plt.xticks(range(transition_prob_min_df.shape[1]), transition_table.var_names, rotation=45, ha='right')
        plt.yticks(range(transition_prob_min_df.shape[0]), transition_table.obs_names)
        plt.xlabel('{} Domains'.format(next_day))
        plt.ylabel('{} Domains'.format(prev_day))
        plt.title('Transition Probability Dotplot')
        plt.colorbar(plt.cm.ScalarMappable(cmap='Reds'), label='Transition Probability', ax=plt.gca())
        plt.tight_layout()
        plt.savefig(args.output_dir+os.sep+f'transition_dotplot_{prev_day}_{next_day}.png')
        plt.close()


import os
import anndata
import torch
from SpaDOT.utils import _train_utils, _utils

import warnings
warnings.filterwarnings("ignore", category=UserWarning, module="torch") # this is for suppressing UserWarning from torch.tensor
 
def train(args):
    # --- load data ---
    print("Loading data...")
    data_dir = os.path.abspath(args.data)
    if not args.output_dir:
        args.output_dir = os.path.dirname(data_dir)

    adata = anndata.read_h5ad(data_dir)
    model_config = _utils.load_model_config(args)
    # add adata related parameters to model_config
    model_config['input_dim'] = adata.n_vars
    tps = adata.obs['timepoint'].unique()
    tps.sort()
    model_config['timepoints'] = tps
    # add device and dtype to model_config
    model_config['device'] = torch.device(args.device)
    model_config['dtype'] = torch.float64

    _utils.set_seed(model_config['seed'])
    # obtain inducing points, graph adjacency matrix, and dataloaders
    print("Preparing data...")
    dataloader_dict = _train_utils.prepare_dataloader(adata, model_config)
    _utils._save_inducing_points(args, dataloader_dict['inducing_points']) # save inducing points

    # train model
    print("Training model...")
    SpaDOT_model, loss_df = _train_utils.train_SpaDOT(dataloader_dict, model_config)
    loss_df.T.to_csv(args.output_dir + os.sep + 'loss.csv')
    if args.save_model:
        torch.save(SpaDOT_model.state_dict(), args.output_dir+os.sep+'SpaDOT_model.pth')
        print("Model saved to %s" % (args.output_dir))
    # obtain the latent representation
    latent_adata = _train_utils.get_latent(SpaDOT_model, model_config, adata, dataloader_dict)
    latent_adata.write_h5ad(args.output_dir+os.sep+args.prefix+'latent.h5ad')

if __name__ == "__main__":
    data_dir = "./examples"
    # create arguments for testing
    class Args:
        data = os.path.join(data_dir, "preprocessed_ChickenHeart.h5ad")
        prefix = ""
        config = None
        save_model = True
        device = 'cuda:0'
    args = Args()
    # create output directory if not exists
    if 'output_dir' not in args.__dict__:
        args.output_dir = os.path.dirname(args.data)
    if not os.path.exists(args.output_dir):
        os.makedirs(args.output_dir)
    # train SpaDOT model
    train(args)


import torch
import torch.nn as nn
import torch.nn.functional as F

class SVGP(nn.Module):
    def __init__(self, model_config, inducing_points, N_train, jitter=1e-2):
        '''
        SVGP model for variational inference in Gaussian processes.
        Parameters
        ----------
        model_config : dict
            Configuration dictionary containing model parameters.
        inducing_points : np.ndarray
            Inducing points for the SVGP model.
        N_train : int
            Number of training points.
        jitter : float
            Jitter term to add to the diagonal of covariance matrices for numerical stability.
        '''
        super(SVGP, self).__init__()
        self.N_train = N_train
        self.jitter = jitter
        # Inducing points
        self.inducing_index_points = torch.tensor(inducing_points, 
                                                  dtype=model_config['dtype']).to(model_config['device'])
        # Kernel
        self.kernel = Kernel(kernel_type=model_config['kernel_type'], 
                             scale=model_config['kernel_scale'], 
                             dtype=model_config['dtype'],
                             device=model_config['device'])


    def _add_diagonal_jitter(self, matrix, jitter):
        """
        Add jitter to the diagonal of a matrix.
        """
        if matrix.dim() == 2:
            Eye = torch.eye(matrix.size(0), device=matrix.device)
        else:
            Eye = torch.eye(matrix.size(-1), device=matrix.device).expand(matrix.shape)
        return matrix + jitter * Eye

    def kernel_matrix(self, x, y, diag_only=False):
        kernel_mat = self.kernel(x, y)
        return torch.diagonal(kernel_mat) if diag_only else kernel_mat

    def variational_loss(self, x, y, noise, mu_hat, A_hat):
        b, m = x.shape[0], self.inducing_index_points.shape[0]
        K_mm = self.kernel_matrix(self.inducing_index_points, self.inducing_index_points)
        K_mm_inv = torch.linalg.inv(self._add_diagonal_jitter(K_mm, self.jitter))

        K_nn = self.kernel_matrix(x, x, diag_only=True)
        K_nm = self.kernel_matrix(x, self.inducing_index_points)
        K_mn = torch.transpose(K_nm, 0, 1)

        mean_vector = torch.matmul(K_nm, torch.matmul(K_mm_inv, mu_hat))
        KL_term = self._compute_kl_term(K_mm, K_mm_inv, mu_hat, A_hat, m)
        L_3_sum_term = self._compute_l3_term(K_nn, K_nm, K_mn, noise, y, mean_vector, K_mm_inv, A_hat, b)

        return L_3_sum_term, KL_term

    def approximate_posterior_params(self, index_points_test, index_points_train, y, noise):
        b = index_points_train.shape[0]
        K_mm = self.kernel_matrix(self.inducing_index_points, self.inducing_index_points)
        K_mm_inv = torch.linalg.inv(self._add_diagonal_jitter(K_mm, self.jitter))

        K_xx = self.kernel_matrix(index_points_test, index_points_test, diag_only=True) 
        K_xm = self.kernel_matrix(index_points_test, self.inducing_index_points)
        K_mx = torch.transpose(K_xm, 0, 1)  # (m, x)

        K_nm = self.kernel_matrix(index_points_train, self.inducing_index_points)
        K_mn = torch.transpose(K_nm, 0, 1) 

        sigma_l = K_mm + (self.N_train / b) * torch.matmul(K_mn, K_nm / noise[:,None])
        sigma_l_inv = torch.linalg.inv(self._add_diagonal_jitter(sigma_l, self.jitter))
        mean_vector = (self.N_train / b) * torch.matmul(K_xm, torch.matmul(sigma_l_inv, torch.matmul(K_mn, y/noise)))

        K_xm_Sigma_l_K_mx = torch.matmul(K_xm, torch.matmul(sigma_l_inv, K_mx))
        B = K_xx + torch.diagonal(-torch.matmul(K_xm, torch.matmul(K_mm_inv, K_mx)) + K_xm_Sigma_l_K_mx)

        mu_hat = (self.N_train / b) * torch.matmul(torch.matmul(K_mm, torch.matmul(sigma_l_inv, K_mn)), y / noise)
        A_hat = torch.matmul(K_mm, torch.matmul(sigma_l_inv, K_mm))

        return mean_vector, B, mu_hat, A_hat

    def _compute_kl_term(self, K_mm, K_mm_inv, mu_hat, A_hat, m):
        K_mm_chol = torch.linalg.cholesky(self._add_diagonal_jitter(K_mm, self.jitter))
        S_chol = torch.linalg.cholesky(self._add_diagonal_jitter(A_hat, self.jitter))
        K_mm_log_det = 2 * torch.sum(torch.log(torch.diagonal(K_mm_chol)))
        S_log_det = 2 * torch.sum(torch.log(torch.diagonal(S_chol)))

        return 0.5 * (K_mm_log_det - S_log_det - m +
                      torch.trace(torch.matmul(K_mm_inv, A_hat)) +
                      torch.sum(mu_hat * torch.matmul(K_mm_inv, mu_hat)))

    def _compute_l3_term(self, K_nn, K_nm, K_mn, noise, y, mean_vector, K_mm_inv, A_hat, b):
        precision = 1 / noise
        K_tilde_terms = precision * (K_nn - torch.diagonal(torch.matmul(K_nm, torch.matmul(K_mm_inv, K_mn))))
        lambda_mat = torch.matmul(K_nm.unsqueeze(2), torch.transpose(K_nm.unsqueeze(2), 1, 2))
        lambda_mat = torch.matmul(K_mm_inv, torch.matmul(lambda_mat, K_mm_inv))
        trace_terms = precision * torch.einsum('bii->b', torch.matmul(A_hat, lambda_mat))
        return -0.5 * (torch.sum(K_tilde_terms) + torch.sum(trace_terms) +
                        torch.sum(torch.log(noise)) + b * torch.log(2 * torch.tensor(torch.pi)) +
                        torch.sum(precision * (y - mean_vector) ** 2))


'''
Kernel same as in SpatialPCA
'''
class Kernel(nn.Module):
    def __init__(self, kernel_type='Gaussian', scale=0.1, dtype=torch.float64, device='cpu'):
        super(Kernel, self).__init__()
        self.kernel_type = kernel_type
        self.scale = torch.tensor([scale], dtype=dtype, device=device)

    def forward(self, x, y):
        d = torch.cdist(x, y, p=2) # Euclidean distance
        # kernel type
        if self.kernel_type == 'Gaussian':
            res = torch.exp(-1*torch.square(d)/self.scale)
        if self.kernel_type == 'Cauchy':
            res = 1 / (1 + 1*torch.square(d)/self.scale)
        if self.kernel_type == 'Quadratic':
            res = 1 - 1 * torch.square(d) / (1 * torch.square(d) + self.scale)
        return res


import ot
import sklearn
import sklearn.metrics
import numpy as np
import torch
from copy import deepcopy
import time

import ctypes
from .ot_func import dummy_c, primal_c, dual_c, compute_duality_gap_c, update_K_c, update_R_c, update_a_b_c, step1_process_c, update_process_c

log_step = False

debug = False
debug_p1 = False
c_for_p1 = True
c_for_v2 = True

default_config = {
    "growth_iters": 3,
    "epsilon": 0.05,
    "lambda1": 1,
    "lambda2": 50,
    "epsilon0": 1,
    "tau": 1000,
    "scaling_iter": 3000,
    "inner_iter_max": 50,
    "tolerance": 1e-8,
    "max_iter": 1e7,
    "batch_size": 5,
    "extra_iter": 1000,
    "numItermax": 1000000,
    "use_Py": False,
    "use_C": True,
    "profiling": False
}  # specified in tutorial


def solve_ot(feats, ot_solver, ot_config, gammas, days, g_est):
    for i in range(len(feats) - 1):
        print("Solving OT for days:", days[i], days[i + 1])
        delta_days = float(days[i + 1]) - float(days[i])
        g = np.power(g_est[i], delta_days)
        gamma = ot_solver(feats[i], feats[i + 1], ot_config, G=g)
        print("Transport matrix shape: ", gamma.shape)
        gammas[f"{i}_{i + 1}"] = gamma


def get_total_ot_loss(feats, indices, gammas):
    # feats (bsz, days, dim)
    days = feats.shape[1]
    alignment_loss = 0
    for i in range(days - 1):
        # for j in range(i + 1, len(feats)):
        feat1 = feats[:, i]
        feat2 = feats[:, i + 1]
        index1 = indices[:, i].numpy()
        index2 = indices[:, i + 1].numpy()
        gamma = gammas[f"{i}_{i+1}"]  # use this to compute alignment loss
        # sample gamma matrix with index here
        gamma = gamma[index1][:, index2]  # sample submatrix of gamma with rna index
        gamma = gamma / gamma.sum(axis=1, keepdims=True)  # normalize rows
        gamma = np.nan_to_num(gamma, nan=0.0, posinf=0.0, neginf=0.0)  # prune invalid values
        gamma = torch.tensor(gamma).cuda().float()
        cost_matrix = torch.cdist(feat1, feat2, p=2)
        transport_cost = torch.mean(gamma * cost_matrix)
        alignment_loss += transport_cost
    alignment_loss /= days - 1
    return alignment_loss


def compute_transport_map_pot(a, b, config, C=None):
    print("Using OT ", config["method"])
    if not isinstance(a, np.ndarray):
        a = a.detach().cpu().numpy()
        b = b.detach().cpu().numpy()
    if C is None:
        C = sklearn.metrics.pairwise.pairwise_distances(a, b, metric="sqeuclidean", n_jobs=-1)
    C = np.ascontiguousarray(C)
    if config["method"] == "emd":
        gamma = ot.emd(ot.unif(a.shape[0]), ot.unif(b.shape[0]), C, numItermax=config["numItermax"])
    elif config["method"] == "sinkhorn":
        gamma = ot.sinkhorn(ot.unif(a.shape[0]), ot.unif(b.shape[0]), C, reg=config["epsilon"])
    elif config["method"] == "unbalanced":
        gamma = ot.unbalanced.sinkhorn_stabilized_unbalanced(
            ot.unif(a.shape[0]),
            ot.unif(b.shape[0]),
            C,
            reg=config["epsilon"],
            reg_m=config["lambda"],
        )  # use wad-OT setting
    return gamma


def compute_transport_map(a, b, config, C=None, G=None):
    if not isinstance(a, np.ndarray):
        a = a.detach().cpu().numpy()
        b = b.detach().cpu().numpy()

    # Compute cost matrix
    if C is None:
        C = sklearn.metrics.pairwise.pairwise_distances(a, b, metric="sqeuclidean", n_jobs=1)
        C = C / np.median(C)

    config["C"] = C  # cost matrix
    if G is None:
        config["G"] = np.ones(C.shape[0])
    else:
        config["G"] = G
    growth_iters = config["growth_iters"]
    gammas = []
    for i in range(growth_iters):
        print("OT iter", i)
        if i == 0:
            row_sums = config["G"]
        else:
            row_sums = gamma.sum(axis=1)  # / tmap.shape[1]
        config["G"] = row_sums
        gamma = optimal_transport_duality_gap(**config)
        gammas.append(deepcopy(gamma))
    return gammas[0]


def fdiv(l, x, p, dx):
    return l * np.sum(dx * (x * (np.log(x / p)) - x + p))


def fdivstar(l, u, p, dx):
    return l * np.sum((p * dx) * (np.exp(u / l) - 1))


def primal(C, K, R, dx, dy, p, q, a, b, epsilon, lambda1, lambda2):
    I = len(p)
    J = len(q)
    F1 = lambda x, y: fdiv(lambda1, x, p, y)
    F2 = lambda x, y: fdiv(lambda2, x, q, y)
    with np.errstate(divide="ignore"):
        return (
            F1(np.dot(R, dy), dx)
            + F2(np.dot(R.T, dx), dy)
            + (epsilon * np.sum(R * np.nan_to_num(np.log(R)) - R + K) + np.sum(R * C)) / (I * J)
        )


def dual(C, K, R, dx, dy, p, q, a, b, epsilon, lambda1, lambda2):
    I = len(p)
    J = len(q)
    F1c = lambda u, v: fdivstar(lambda1, u, p, v)
    F2c = lambda u, v: fdivstar(lambda2, u, q, v)

    t1 = -F1c(-epsilon * np.log(a), dx)
    t2 = -F2c(-epsilon * np.log(b), dy)
    t3 = -epsilon * np.sum(R - K) / (I * J)

    #print("py", I, J, lambda1, lambda2, t1, t2, t3)
    #return -F1c(-epsilon * np.log(a), dx) - F2c(-epsilon * np.log(b), dy) - epsilon * np.sum(R - K) / (I * J)

    return t1 + t2 + t3


# end @ Lénaïc Chizat


def optimal_transport_duality_gap(
    C,
    G,
    lambda1,
    lambda2,
    epsilon,
    batch_size,
    tolerance,
    tau,
    epsilon0,
    max_iter,
    use_Py,
    use_C,
    profiling,
    **ignored,
):
    """
    Compute the optimal transport with stabilized numerics, with the guarantee that the duality gap is at most `tolerance`
    Parameters
    ----------
    C : 2-D ndarray
        The cost matrix. C[i][j] is the cost to transport cell i to cell j
    G : 1-D array_like
        Growth value for input cells.
    lambda1 : float, optional
        Regularization parameter for the marginal constraint on p
    lambda2 : float, optional
        Regularization parameter for the marginal constraint on q
    epsilon : float, optional
        Entropy regularization parameter.
    batch_size : int, optional
        Number of iterations to perform between each duality gap check
    tolerance : float, optional
        Upper bound on the duality gap that the resulting transport map must guarantee.
    tau : float, optional
        Threshold at which to perform numerical stabilization
    epsilon0 : float, optional
        Starting value for exponentially-decreasing epsilon
    max_iter : int, optional
        Maximum number of iterations. Print a warning and return if it is reached, even without convergence.
    Returns
    -------
    transport_map : 2-D ndarray
        The entropy-regularized unbalanced transport map
    """
    cnt_step0 = 0
    cnt_step1 = 0
    cnt_step2 = 0

    time_i0 = time.time()

    C = np.asarray(C, dtype=np.float64)

    epsilon_scalings = 5
    scale_factor = np.exp(-np.log(epsilon) / epsilon_scalings)

    I, J = C.shape
    dx, dy = np.ones(I) / I, np.ones(J) / J

    p = G
    q = np.ones(C.shape[1]) * np.average(G)

    u, v = np.zeros(I), np.zeros(J)
    a, b = np.ones(I), np.ones(J)

    dtype = np.float64
    C  = np.ascontiguousarray(C, dtype=dtype)
    dx = np.ascontiguousarray(dx, dtype=dtype)
    dy = np.ascontiguousarray(dy, dtype=dtype)
    p  = np.ascontiguousarray(p, dtype=dtype)
    q  = np.ascontiguousarray(q, dtype=dtype)
    u  = np.ascontiguousarray(u, dtype=dtype)
    v  = np.ascontiguousarray(v, dtype=dtype)
    a  = np.ascontiguousarray(a, dtype=dtype)
    b  = np.ascontiguousarray(b, dtype=dtype)

    epsilon_i = epsilon0 * scale_factor
    current_iter = 0

    for e in range(epsilon_scalings + 1):
        if profiling:
            print('Compute for epsilon scaling {}'.format(e))

        duality_gap = np.inf

        u = u + epsilon_i * np.log(a)
        v = v + epsilon_i * np.log(b)  # absorb

        a, b = np.ones(I), np.ones(J)

        epsilon_i = epsilon_i / scale_factor

        alpha1 = lambda1 / (lambda1 + epsilon_i)
        alpha2 = lambda2 / (lambda2 + epsilon_i)

        old_a = a.copy()
        old_b = b.copy()

        threshold = tolerance if e == epsilon_scalings else 1e-6

        time_k0 = time.time()

        if use_C:
            K = np.zeros_like(C)
            _K = np.zeros_like(C)
            update_K_c(K, _K, C, u, v, epsilon_i)
        else:
            _K = np.exp(-C / epsilon_i)
            K = np.exp((np.array([u]).T - C + np.array([v])) / epsilon_i)

        time_k5 = time.time()

        R = np.zeros(K.shape)

        if profiling:
            print('step0 {:3f} ms'.format(1000 * (time_k5 - time_k0)))
        cnt_step0 += time_k5 - time_k0

        if use_C and c_for_v2:
            duality_gap = update_process_c(
                R, a, b, old_a, old_b, K, _K, C,
                dx, dy, p, q, u, v,
                epsilon_scalings, e, batch_size, epsilon_i, threshold,
                tau, lambda1, lambda2, alpha1, alpha2,
                current_iter, max_iter
            )
        else:
            while duality_gap > threshold:
                if debug:
                    print('  {} - duality_gap {}'.format(current_iter, duality_gap))

                if debug_p1:
                    print('pre', a.sum(), b.sum())

                if not c_for_p1 or use_Py:
                    if debug_p1:
                        print('a', a.sum())
                        print('b', b.sum())

                    time_s1 = time.time()
                    for i in range(batch_size if e == epsilon_scalings else 5):
                        current_iter += 1
                        old_a, old_b = a, b

                        if profiling:
                            time_h1 = time.time()

                        a = (p / (K.dot(np.multiply(b, dy)))) ** alpha1 * np.exp(-u / (lambda1 + epsilon_i))

                        if profiling:
                            time_h2 = time.time()

                        b = (q / (K.T.dot(np.multiply(a, dx)))) ** alpha2 * np.exp(-v / (lambda2 + epsilon_i))

                        if profiling:
                            time_h3 = time.time()

                        if profiling:
                            print('step1 {:3f} ms'.format(1000 * (time_h2 - time_h1)))
                            print('step1 {:3f} ms'.format(1000 * (time_h3 - time_h2)))

                        if debug_p1:
                            print('a', a.sum())
                            print('b', b.sum())

                        # stabilization
                        if max(max(abs(a)), max(abs(b))) > tau:
                            u = u + epsilon_i * np.log(a)
                            v = v + epsilon_i * np.log(b)  # absorb
                            K = np.exp((np.array([u]).T - C + np.array([v])) / epsilon_i)
                            a, b = np.ones(I), np.ones(J)

                            if debug_p1:
                                print('stabilization')

                        if current_iter >= max_iter:
                            logger.warning("Reached max_iter with duality gap still above threshold. Returning")
                            return (K.T * a).T * b

                    time_s2 = time.time()

                if c_for_p1 and use_C:
                    time_s1 = time.time()
                    current_iter = step1_process_c(
                        a, b, old_a, old_b, K, C, dx, dy, p, q, u, v,
                        current_iter, max_iter, batch_size if e == epsilon_scalings else 5,
                        tau, lambda1, lambda2, alpha1, alpha2, epsilon_i)
                    time_s2 = time.time()

                if debug_p1:
                    print('post', a.sum(), b.sum())
                    print('old ab', old_a.sum(), old_b.sum())
                    print('uv', u.sum(), v.sum())

                if profiling:
                    print('step1 {:3f} ms'.format(1000 * (time_s2 - time_s1)))
                cnt_step1 += time_s2 - time_s1

                time_s3 = time.time()

                # The real dual variables. a and b are only the stabilized variables
                _a = a * np.exp(u / epsilon_i)
                _b = b * np.exp(v / epsilon_i)

                # Skip duality gap computation for the first epsilon scalings, use dual variables evolution instead
                if e == epsilon_scalings:
                    R_c = np.zeros(K.shape)

                    if use_Py:
                        time_d5 = time.time()
                        R = (K.T * a).T * b
                        time_d6 = time.time()

                    if use_C:
                        time_d7 = time.time()
                        update_R_c(R_c, K, a, b)
                        time_d8 = time.time()

                    if use_Py and use_C:
                        if profiling:
                            print("  update_R      :            cost : {:5f} ms".format(1000 * (time_d6 - time_d5)))
                            print("  update_R_c    :            cost : {:5f} ms".format(1000 * (time_d8 - time_d7)))

                        assert(np.allclose(R, R_c))

                    if not use_Py and use_C:
                        if profiling:
                            print("  update_R_c    :            cost : {:5f} ms".format(1000 * (time_d8 - time_d7)))
                        R = R_c

                    if use_Py:
                        time_d1 = time.time()
                        pri = primal(C, _K, R, dx, dy, p, q, _a, _b, epsilon_i, lambda1, lambda2)
                        dua = dual(C, _K, R, dx, dy, p, q, _a, _b, epsilon_i, lambda1, lambda2)
                        time_d2 = time.time()

                        duality_gap = (pri - dua) / abs(pri)

                    if use_C:
                        time_d3 = time.time()
                        duality_gap_c = compute_duality_gap_c(C, _K, R, dx, dy, p, q, _a, _b, epsilon_i, lambda1, lambda2)
                        time_d4 = time.time()

                    if use_Py and use_C:
                        if profiling:
                            print("  duality_gap   : {:5f}   cost : {:5f} ms".format(duality_gap, 1000 * (time_d2 - time_d1)))
                            print("  duality_gap_c : {:5f}   cost : {:5f} ms".format(duality_gap_c, 1000 * (time_d4 - time_d3)))

                        assert(np.allclose(duality_gap, duality_gap_c))

                    if not use_Py and use_C:
                        duality_gap = duality_gap_c

                        if profiling:
                            print("  duality_gap_c : {:5f}   cost : {:5f} ms".format(duality_gap_c, 1000 * (time_d4 - time_d3)))

                else:
                    duality_gap = max(
                        np.linalg.norm(_a - old_a * np.exp(u / epsilon_i)) / (1 + np.linalg.norm(_a)),
                        np.linalg.norm(_b - old_b * np.exp(v / epsilon_i)) / (1 + np.linalg.norm(_b)),
                    )

                time_s4 = time.time()

                if profiling:
                    print('step2 {:3f} ms'.format(1000 * (time_s4 - time_s3)))
                cnt_step2 += time_s4 - time_s3

                if debug_p1:
                    print('last', a.sum(), b.sum())

    time_i1 = time.time()

    if log_step:
        print('---------------------------------------------------------')
        print('Step0 {:5f} ms'.format(1000 * cnt_step0))
        print('Step1 {:5f} ms'.format(1000 * cnt_step1))
        print('Step2 {:5f} ms'.format(1000 * cnt_step2))
        print('Total {:5f} ms'.format(1000 * (time_i1 - time_i0)))
        print('DualityGap', duality_gap, 'R.sum', R.sum())
        print('---------------------------------------------------------')

    if np.isnan(duality_gap):
        raise RuntimeError("Overflow encountered in duality gap computation, please report this incident")

    return R / C.shape[1]


def transport_stablev2(
    C,
    lambda1,
    lambda2,
    epsilon,
    scaling_iter,
    G,
    tau,
    epsilon0,
    extra_iter,
    inner_iter_max,
    **ignored,
):
    """
    Compute the optimal transport with stabilized numerics.
    Args:
        C: cost matrix to transport cell i to cell j
        lambda1: regularization parameter for marginal constraint for p.
        lambda2: regularization parameter for marginal constraint for q.
        epsilon: entropy parameter
        scaling_iter: number of scaling iterations
        G: growth value for input cells
    """

    warm_start = tau is not None
    epsilon_final = epsilon

    def get_reg(n):  # exponential decreasing
        return (epsilon0 - epsilon_final) * np.exp(-n) + epsilon_final

    epsilon_i = epsilon0 if warm_start else epsilon
    dx = np.ones(C.shape[0]) / C.shape[0]
    dy = np.ones(C.shape[1]) / C.shape[1]

    p = G
    q = np.ones(C.shape[1]) * np.average(G)

    u = np.zeros(len(p))
    v = np.zeros(len(q))
    b = np.ones(len(q))
    K = np.exp(-C / epsilon_i)

    alpha1 = lambda1 / (lambda1 + epsilon_i)
    alpha2 = lambda2 / (lambda2 + epsilon_i)
    epsilon_index = 0
    iterations_since_epsilon_adjusted = 0
    eps = 1e-10
    for i in range(scaling_iter):
        # scaling iteration
        a = (p / (K.dot(np.multiply(b, dy)) + eps)) ** alpha1 * np.exp(-u / (lambda1 + epsilon_i))
        b = (q / (K.T.dot(np.multiply(a, dx)) + eps)) ** alpha2 * np.exp(-v / (lambda2 + epsilon_i))

        # stabilization
        iterations_since_epsilon_adjusted += 1
        if max(max(abs(a)), max(abs(b))) > tau:
            u = u + epsilon_i * np.log(a)
            v = v + epsilon_i * np.log(b)  # absorb
            K = np.exp((np.array([u]).T - C + np.array([v])) / epsilon_i)
            a = np.ones(len(p))
            b = np.ones(len(q))

        if warm_start and iterations_since_epsilon_adjusted == inner_iter_max:
            epsilon_index += 1
            iterations_since_epsilon_adjusted = 0
            u = u + epsilon_i * np.log(a)
            v = v + epsilon_i * np.log(b)  # absorb
            epsilon_i = get_reg(epsilon_index)
            alpha1 = lambda1 / (lambda1 + epsilon_i)
            alpha2 = lambda2 / (lambda2 + epsilon_i)
            K = np.exp((np.array([u]).T - C + np.array([v])) / epsilon_i)
            a = np.ones(len(p))
            b = np.ones(len(q))

    for i in range(extra_iter):
        a = (p / (K.dot(np.multiply(b, dy)) + eps)) ** alpha1 * np.exp(-u / (lambda1 + epsilon_i))
        b = (q / (K.T.dot(np.multiply(a, dx)) + eps)) ** alpha2 * np.exp(-v / (lambda2 + epsilon_i))

    R = (K.T * a).T * b

    return R / C.shape[1]



import os, sys, argparse

def parse_args():
    parser = argparse.ArgumentParser(description="SpaDOT: Optimal transport modeling uncovers spatial domain dynamics in spatiotemporal transcriptomics.",
        prog='SpaDOT')
    subparsers = parser.add_subparsers(help='sub-command help.', dest='cmd_choice')

    # ===========================
    # Parser for performing preprocess for spatiotemporal transcriptomics data
    # -i, --data
    # -o, --output_dir
    # --prefix
    # --feature_selection
    # ============================
    preprocess_parser = subparsers.add_parser('preprocess', 
            help='(Recommended but optional) Perform data preprocessing and feature selection.')
    preprocess_parser.add_argument('-i', '--data', dest='data',
            required=True, type=str, help="We preprocess the data ")
    preprocess_parser.add_argument('-o', '--output_dir', dest='output_dir',
            type=str, help="Output directory to store preprocessed data. Default: the same as where the data locates.")
    preprocess_parser.add_argument('--prefix', dest='prefix',
            type=str, default='preprocessed_', help="Prefix for preprocessed data. Default: preprocessed_")
    preprocess_parser.add_argument('--feature_selection', dest='feature_selection',
            default=True, action="store_false",
            help="Whether performing feature selection. If not specified, feature selection option is chosen.")

    # ===========================
    # Parser for training SpaDOT model to obtain latent representations
    # -i, --data
    # -o, --output_dir
    # --prefix
    # --config
    # --save_model
    # --device
    # ============================
    train_parser = subparsers.add_parser('train', help='Train a SpaDOT model.')
    train_parser.add_argument('-i', '--data', dest='data', 
            type=str, help="A anndata object containing time point information and spatial coordinates.")
    train_parser.add_argument('-o', '--output_dir', dest='output_dir', 
            type=str, help="Output directory to store latent representations. Default: the same as where the data locates.")
    train_parser.add_argument('--prefix', dest='prefix', 
            type=str, default='', help="Prefix for output latent representations. Default: ''")
    train_parser.add_argument('--config', dest='config',
            type=str, help="Path to the config file, in a yaml format.")
    train_parser.add_argument('--device', dest='device',
            type=str, default='cuda:0', help="Device to use for training. Default: cuda:0")
    train_parser.add_argument('--save_model', dest='save_model',
            default=False, action="store_true",
            help="Whether saving the trained model. If specified, the trained model will be stored in the output_dir as 'model.pt'.")

    ## ===========================
    # Parser for analyzing latent representations
    # -i, --data
    # -o, --output_dir
    # --prefix
    # ============================
    analyze_parser = subparsers.add_parser('analyze', help='Analyze the latent representations generated by SpaDOT model.')
    analyze_parser.add_argument('-i', '--data', dest='data', 
            type=str, required=True,
            help="An anndata object containing latent representations.")
    analyze_parser.add_argument('-o', '--output_dir', dest='output_dir', 
            type=str, help="Output directory to store analyses results. Default: the same as where the data locates.")
    analyze_parser.add_argument('--prefix', dest='prefix', 
            type=str, default='', help="Prefix for output analyses results. Default: ''")
    analyze_parser.add_argument('--n_clusters', dest='n_clusters',
                                type=lambda s: [int(item) for item in s.split(',')],
                                help="A comma-separated list of integers representing the number of clusters for analysis.")
    args = parser.parse_args()
    print("User input arguments: ", args)
    return args


def main():
    args = parse_args()
    ## preprocessing
    if "preprocess" == args.cmd_choice:
        from SpaDOT import preprocess
        if not os.path.exists(args.data):
            sys.exit("The data does not exist! Please make sure the data is correctly specified.")
        preprocess(args)
        
    ## train to obtain latent representations
    if "train" == args.cmd_choice:
        from SpaDOT import train
        if not os.path.exists(args.data):
            sys.exit("The preprocessed data does not exist! Please make sure the data is correctly specified.")
        train(args)

    ## analyze latent representations
    if "analyze" == args.cmd_choice:
        from SpaDOT import analyze
        if not os.path.exists(args.data):
            sys.exit("The latent representations does not exist! Please make sure the latent representations are correctly generated from the last step.")
        analyze(args)

if __name__ == "__main__":
    main()


import os
import random
import numpy as np
import anndata
from SpaDOT.utils import _preprocess_utils

seed=1993
random.seed(seed)

def preprocess(args):
    '''
    Preprocess the data for SpaDOT model.
    '''
    data_dir = os.path.abspath(args.data)
    if not args.output_dir:
        args.output_dir = os.path.dirname(data_dir)
    adata = anndata.read_h5ad(data_dir)

    # check if `timepoint` in adata.obs
    if 'timepoint' not in adata.obs.columns:
        raise ValueError("The `timepoint` column is not found in adata.obs. Please make sure timepoint information is given.")
    if 'spatial' not in adata.obsm.keys():
        raise ValueError("The `spatial` key is not found in adata.obsm. Please make sure spatial coordinates are provided.")
    # check if adata.obsm['spatial'] is 2D numpy array
    if not isinstance(adata.obsm['spatial'], np.ndarray) or adata.obsm['spatial'].ndim != 2:
        raise ValueError("The `spatial` key in adata.obsm is not a 2D numpy array. Please make sure spatial coordinates are correctly provided.")
    # preprocess adata based on whether performing SVG selection
    preprocessed_adata = _preprocess_utils.preprocess_adata(args, adata)
    preprocessed_adata.write_h5ad(os.path.join(args.output_dir, args.prefix + os.path.basename(data_dir)))

if __name__ == '__main__':
    data_dir = "./examples"
    # create arguments for testing
    class Args:
        data = os.path.join(data_dir, "ChickenHeart.h5ad")
        prefix = "preprocessed_"
        feature_selection = True
    args = Args()
    # create output directory if not exists
    if 'output_dir' not in args.__dict__:
        args.output_dir = os.path.dirname(args.data)
    if not os.path.exists(args.output_dir):
        os.makedirs(args.output_dir)
    # run preprocessing
    preprocess(args)


import torch
import torch.nn as nn
import torch.nn.functional as F
from torch_geometric.nn import GATConv


class SVGPEncoder(nn.Module):
    def __init__(self, input_dim, SVGP_z_dim, hidden_dims):
        super(SVGPEncoder, self).__init__()
        # Create a sequential network with specified hidden dimensions
        layers = [input_dim] + hidden_dims
        SVGP_encoder_net = []
        for i in range(1, len(layers)):
            linear_layer = nn.Linear(layers[i-1], layers[i])
            nn.init.xavier_uniform_(linear_layer.weight)
            SVGP_encoder_net.append(linear_layer)
            # normalization
            SVGP_encoder_net.append(nn.BatchNorm1d(layers[i]))
            # activation
            SVGP_encoder_net.append(nn.LeakyReLU())
        self.SVGP_encoder_net = nn.Sequential(*SVGP_encoder_net)
        self.SVGP_fc = nn.Linear(hidden_dims[-1], SVGP_z_dim*2)
        nn.init.xavier_uniform_(self.SVGP_fc.weight)


    def forward(self, x):
        '''
        x: gene expression
        '''
        h = self.SVGP_encoder_net(x)
        # Compute mean and log variance for the latent space
        SVGP_z = self.SVGP_fc(h)
        SVGP_enc_mu, SVGP_enc_logvar = torch.chunk(SVGP_z, 2, dim=1)
        return SVGP_enc_mu, torch.exp(SVGP_enc_logvar)  # Return mean and variance

# Define the GATEncoder class, which uses Graph Attention Networks (GAT) for encoding graph data
class GATEncoder(nn.Module):
    def __init__(self, input_dim, GAT_z_dim, hidden_dim=512, num_heads=4):
        super(GATEncoder, self).__init__()
        # Define multiple GAT layers with specified dimensions and heads
        self.gat1 = GATConv(input_dim, hidden_dim, heads=num_heads, concat=True)
        nn.init.xavier_uniform(self.gat1.lin.weight)
        self.gat2 = GATConv(hidden_dim * num_heads, hidden_dim, heads=num_heads, concat=True)
        nn.init.xavier_uniform(self.gat2.lin.weight)
        self.gat3 = GATConv(hidden_dim * num_heads, hidden_dim, heads=num_heads, concat=False)
        nn.init.xavier_uniform(self.gat3.lin.weight)
        self.GAT_fc = nn.Linear(hidden_dim, GAT_z_dim*2)
        nn.init.xavier_uniform_(self.GAT_fc.weight)

    def forward(self, x, edge_index):
        '''
        x: gene expression
        edge_index: graph structure
        '''
        # Pass input through each GAT layer with LeakyReLU activation
        h = F.leaky_relu(self.gat1(x, edge_index))
        h = F.leaky_relu(self.gat2(h, edge_index))
        h = self.gat3(h, edge_index)
        GAT_z = self.GAT_fc(h)
        GAT_enc_mu, GAT_enc_logvar = torch.chunk(GAT_z, 2, dim=1)
        return GAT_enc_mu, torch.exp(GAT_enc_logvar)




import torch.nn as nn

class Decoder(nn.Module):
    def __init__(self, input_dim, z_dim, decoder_layers):
        super(Decoder, self).__init__()
        layers = [z_dim] + decoder_layers + [input_dim]
        decoder_net = []
        for i in range(1, len(layers)-1):
            linear_layer = nn.Linear(layers[i-1], layers[i])
            nn.init.xavier_uniform_(linear_layer.weight)
            decoder_net.append(linear_layer)
            # normalization
            decoder_net.append(nn.LayerNorm(layers[i]))
            # activation
            decoder_net.append(nn.LeakyReLU())
        decoder_net.append(nn.Linear(layers[-2], layers[-1]))
        self.decoder_net = nn.Sequential(*decoder_net)

    def forward(self, latent_sample):
        return self.decoder_net(latent_sample)



from setuptools import setup, find_packages

setup(
    name="SpaDOT",
    version="1.0.0-beta",
    description="Package for paper: Optimal transport modeling uncovers spatial domain dynamics in spatiotemporal transcriptomics",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Wenjing Ma",
    author_email="mawenjing1993@gmail.com",
    url="https://http://marvinquiet.github.io/SpaDOT",
    packages=find_packages(),
    package_data={
        "SpaDOT.utils.OT_loss": ["libot.so"],  # Include the .so file
        "SpaDOT": ["config.yaml"] # Include the config.yaml file
    },
    install_requires=[
        "torch==2.5.0",
        "torchvision",
        "torchaudio",
        # "torch_geometric==2.6.1", # install after pyg-lib, torch-sparse and torch-scatter
        "anndata==0.9.1",
        "scanpy==1.9.8",
        "numpy<2.0.0", # compatible with scanpy 1.9.8
        "wot", # Optimal transport library
        "pandas",
        "scipy", 
        "scikit-learn",  # sklearn is part of scikit-learn
        "matplotlib",
        "seaborn",
        "tqdm",
        "pyyaml",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9",
    entry_points={
        "console_scripts": [
            "SpaDOT=SpaDOT.cli:main",  # Expose the CLI via the `SpaDOT` command
        ],
    },
)



import os
import scipy
import anndata
import scanpy as sc
import pandas as pd

def preprocess_adata(args, adata):
    '''
    Preprocess the data for SpaDOT model training.
    Args:
        args: arguments
        adata: AnnData object
    Returns:
        preprocessed_adata: preprocessed AnnData object
    '''
    tps = adata.obs['timepoint'].unique()
    if args.feature_selection:
        SVGs = _get_SVGs(args, tps)
        adata = adata[:, SVGs].copy()
    # save counts in adata.layers
    if not scipy.sparse.issparse(adata.X):
        adata.layers['counts'] = scipy.sparse.csr_matrix(adata.X)
    else:
        adata.layers['counts'] = adata.X
    # --- preprocess data
    tp_adata_list = []
    for tp in tps:
        tp_adata = adata[adata.obs['timepoint'] == tp]
        sc.pp.normalize_total(tp_adata, target_sum=1e-4)
        sc.pp.log1p(tp_adata)
        tp_adata_list.append(tp_adata)
    # union SVG genes for all time points
    if args.feature_selection:
        genes = set().union(*(obj.var_names for obj in tp_adata_list))
        genes = list(genes)
        genes.sort()
        with open(args.output_dir+os.sep+'SVG_genes.txt', 'w') as f:
            for item in genes:
                f.write("%s\n" % item)  
    new_tp_adata_list = []
    for tp_adata in tp_adata_list:
        if args.feature_selection:
            tp_adata = tp_adata[:, genes].copy()
        sc.pp.scale(tp_adata)
        new_tp_adata_list.append(tp_adata)
    preprocessed_adata = anndata.concat(new_tp_adata_list)
    return preprocessed_adata

def _get_SVGs(args, tps):
    '''
    Run SPARKX to select SVGs.
    Args:
        args: arguments
    Returns:
        SVG_genes: list of SVG genes
    '''
    # --- after running SPARKX, the SVGs are saved in the result directory
    ## select SVGs by cluster
    tp_SVGs = []
    for tp in tps:
        tp_SVGs_cluster = pd.read_csv(args.output_dir+os.sep+str(tp)+'_SVG_sparkx_clustered_louvain.csv', header=0, index_col=0)
        tp_SVGs.append(tp_SVGs_cluster)
    min_idx = min(range(len(tp_SVGs)), key=lambda i: len(tp_SVGs[i]))
    min_tp_SVG_len = len(tp_SVGs[min_idx])
    SVG_genes = tp_SVGs[min_idx].index.tolist()
    for idx, tp_SVG in enumerate(tp_SVGs):
        if idx == min_idx: continue
        tp_SVG_num_clusters = len(set(tp_SVG['cluster']))
        tp_SVG_top = tp_SVG.sort_values(by='adjustedPval', ascending=True).groupby('cluster').head(max(100, round(min_tp_SVG_len/tp_SVG_num_clusters)))
        SVG_genes.extend(tp_SVG_top.index.tolist())
    SVG_genes = list(set(SVG_genes))
    SVG_genes.sort()
    return SVG_genes


# --- other utilitities functions ---
import os
import yaml
# from importlib.resources import files
import sklearn
import scipy as sp
import numpy as np
import pandas as pd
import torch
import random
from torch.backends import cudnn

def set_seed(seed=1993):
    os.environ['PYTHONHASHSEED'] = str(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)
    random.seed(seed)
    cudnn.deterministic = True
    cudnn.benchmark = False
    g = torch.Generator()
    g.manual_seed(seed) # for stabilization

def seed_worker(worker_id):
    np.random.seed(1993)
    random.seed(1993)

def load_model_config(args):
    '''
    Load the model configuration from the config.yaml file provided in the package.
    '''
    if args.config:
        config_path = args.config
    else:
        # Load the default config.yaml from the package
        # config_path = files("SpaDOT").joinpath("config.yaml")
        config_path = os.path.join(os.path.dirname(os.path.dirname(__file__)), "config.yaml")
    with open(config_path, 'r') as f:
        config = yaml.safe_load(f)
    return config

def _Cal_Spatial_Net(adata, k_cutoff=None, max_neigh=30):
    """
    Construct the spatial neighbor networks using KNN.
    Parameters

    adata : anndata.AnnData
        The AnnData object containing the spatial data.
    """

    print('Calculating spatial graph...')
    coor = pd.DataFrame(adata.obsm['spatial'])
    coor.index = adata.obs.index
    coor.columns = ['imagerow', 'imagecol']

    nbrs = sklearn.neighbors.NearestNeighbors(
        n_neighbors=max_neigh + 1, algorithm='auto').fit(coor)
    distances, indices = nbrs.kneighbors(coor)
    indices = indices[:, 1:k_cutoff + 1]
    distances = distances[:, 1:k_cutoff + 1]

    KNN_list = []
    for it in range(indices.shape[0]):
        KNN_list.append(pd.DataFrame(zip([it] * indices.shape[1], indices[it, :], distances[it, :])))
    KNN_df = pd.concat(KNN_list)
    KNN_df.columns = ['Cell1', 'Cell2', 'Distance']

    Spatial_Net = KNN_df.copy()
    id_cell_trans = dict(zip(range(coor.shape[0]), np.array(coor.index), ))
    Spatial_Net['Cell1'] = Spatial_Net['Cell1'].map(id_cell_trans)
    Spatial_Net['Cell2'] = Spatial_Net['Cell2'].map(id_cell_trans)

    print('The graph contains %d edges, %d cells.' % (Spatial_Net.shape[0], adata.n_obs))
    print('%.4f neighbors per cell on average.' % (Spatial_Net.shape[0] / adata.n_obs))
    adata.uns['Spatial_Net'] = Spatial_Net

    X = pd.DataFrame(adata.layers['counts'].toarray()[:, ], index=adata.obs.index, columns=adata.var.index)
    cells = np.array(X.index)
    cells_id_tran = dict(zip(cells, range(cells.shape[0])))
    if 'Spatial_Net' not in adata.uns.keys():
        raise ValueError("Spatial_Net is not existed! Run Cal_Spatial_Net first!")
        
    # add self-loop
    Spatial_Net = adata.uns['Spatial_Net']
    G_df = Spatial_Net.copy()
    G_df['Cell1'] = G_df['Cell1'].map(cells_id_tran)
    G_df['Cell2'] = G_df['Cell2'].map(cells_id_tran)
    G = sp.sparse.coo_matrix((np.ones(G_df.shape[0]), (G_df['Cell1'], G_df['Cell2'])), shape=(adata.n_obs, adata.n_obs))
    G = G + np.eye(G.shape[0])  # add self-loop
    adata.uns['adj'] = G

def _save_inducing_points(args, inducing_points_dict):
    """
    Write the inducing points to a file.
    """
    # Convert the dictionary into a DataFrame directly
    inducing_points_list = []
    for key, value in inducing_points_dict.items():
        # Create a DataFrame for each key-value pair
        df = pd.DataFrame(value)
        df.columns = ['norm-pixel_x', 'norm-pixel_y']
        df["timepoint"] = key  # Add the key as a new column
        inducing_points_list.append(df)
    # Combine all DataFrames into a single DataFrame
    inducing_points_df = pd.concat(inducing_points_list, ignore_index=True)

    # Save to a file
    inducing_points_df.to_csv(args.output_dir+os.sep+args.prefix+"inducing_points.csv", index=False)


import os
import math
import ctypes

import numpy as np
from numpy.ctypeslib import ndpointer

cur_folder = os.path.dirname(os.path.abspath(__file__))

lib = ctypes.cdll.LoadLibrary(os.path.join(cur_folder, 'libot.so'))

lib.dummy_float.argtypes = [
    ndpointer(dtype=ctypes.c_float, ndim=2, flags='C_CONTIGUOUS'), # C
    ndpointer(dtype=ctypes.c_float, ndim=2, flags='C_CONTIGUOUS'), # K
    ndpointer(dtype=ctypes.c_float, ndim=2, flags='C_CONTIGUOUS'), # R
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # dx
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # dy
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # p
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # q
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # a
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # b
    ctypes.c_float, # epsilon
    ctypes.c_float, # lambda1
    ctypes.c_float, # lambda2
    ctypes.c_int, # m
    ctypes.c_int  # n
]

lib.dummy_float.restype = ctypes.c_float

lib.dummy_double.argtypes = [
    ndpointer(dtype=ctypes.c_double, ndim=2, flags='C_CONTIGUOUS'), # C
    ndpointer(dtype=ctypes.c_double, ndim=2, flags='C_CONTIGUOUS'), # K
    ndpointer(dtype=ctypes.c_double, ndim=2, flags='C_CONTIGUOUS'), # R
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # dx
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # dy
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # p
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # q
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # a
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # b
    ctypes.c_double, # epsilon
    ctypes.c_double, # lambda1
    ctypes.c_double, # lambda2
    ctypes.c_int, # m
    ctypes.c_int  # n
]

lib.dummy_double.restype = ctypes.c_double

lib.primal_float.argtypes = [
    ndpointer(dtype=ctypes.c_float, ndim=2, flags='C_CONTIGUOUS'), # C
    ndpointer(dtype=ctypes.c_float, ndim=2, flags='C_CONTIGUOUS'), # K
    ndpointer(dtype=ctypes.c_float, ndim=2, flags='C_CONTIGUOUS'), # R
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # dx
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # dy
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # p
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # q
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # a
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # b
    ctypes.c_float, # epsilon
    ctypes.c_float, # lambda1
    ctypes.c_float, # lambda2
    ctypes.c_int, # m
    ctypes.c_int  # n
]

lib.primal_float.restype = ctypes.c_float

lib.primal_double.argtypes = [
    ndpointer(dtype=ctypes.c_double, ndim=2, flags='C_CONTIGUOUS'), # C
    ndpointer(dtype=ctypes.c_double, ndim=2, flags='C_CONTIGUOUS'), # K
    ndpointer(dtype=ctypes.c_double, ndim=2, flags='C_CONTIGUOUS'), # R
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # dx
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # dy
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # p
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # q
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # a
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # b
    ctypes.c_double, # epsilon
    ctypes.c_double, # lambda1
    ctypes.c_double, # lambda2
    ctypes.c_int, # m
    ctypes.c_int  # n
]

lib.primal_double.restype = ctypes.c_double

lib.dual_float.argtypes = [
    ndpointer(dtype=ctypes.c_float, ndim=2, flags='C_CONTIGUOUS'), # C
    ndpointer(dtype=ctypes.c_float, ndim=2, flags='C_CONTIGUOUS'), # K
    ndpointer(dtype=ctypes.c_float, ndim=2, flags='C_CONTIGUOUS'), # R
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # dx
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # dy
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # p
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # q
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # a
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # b
    ctypes.c_float, # epsilon
    ctypes.c_float, # lambda1
    ctypes.c_float, # lambda2
    ctypes.c_int, # m
    ctypes.c_int  # n
]

lib.dual_float.restype = ctypes.c_float

lib.dual_double.argtypes = [
    ndpointer(dtype=ctypes.c_double, ndim=2, flags='C_CONTIGUOUS'), # C
    ndpointer(dtype=ctypes.c_double, ndim=2, flags='C_CONTIGUOUS'), # K
    ndpointer(dtype=ctypes.c_double, ndim=2, flags='C_CONTIGUOUS'), # R
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # dx
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # dy
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # p
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # q
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # a
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # b
    ctypes.c_double, # epsilon
    ctypes.c_double, # lambda1
    ctypes.c_double, # lambda2
    ctypes.c_int, # m
    ctypes.c_int  # n
]

lib.dual_double.restype = ctypes.c_double

lib.compute_duality_gap_float.argtypes = [
    ndpointer(dtype=ctypes.c_float, ndim=2, flags='C_CONTIGUOUS'), # C
    ndpointer(dtype=ctypes.c_float, ndim=2, flags='C_CONTIGUOUS'), # K
    ndpointer(dtype=ctypes.c_float, ndim=2, flags='C_CONTIGUOUS'), # R
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # dx
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # dy
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # p
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # q
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # a
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # b
    ctypes.c_float, # epsilon
    ctypes.c_float, # lambda1
    ctypes.c_float, # lambda2
    ctypes.c_int, # m
    ctypes.c_int  # n
]

lib.compute_duality_gap_float.restype = ctypes.c_float

lib.compute_duality_gap_double.argtypes = [
    ndpointer(dtype=ctypes.c_double, ndim=2, flags='C_CONTIGUOUS'), # C
    ndpointer(dtype=ctypes.c_double, ndim=2, flags='C_CONTIGUOUS'), # K
    ndpointer(dtype=ctypes.c_double, ndim=2, flags='C_CONTIGUOUS'), # R
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # dx
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # dy
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # p
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # q
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # a
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # b
    ctypes.c_double, # epsilon
    ctypes.c_double, # lambda1
    ctypes.c_double, # lambda2
    ctypes.c_int, # m
    ctypes.c_int  # n
]

lib.compute_duality_gap_double.restype = ctypes.c_double

lib.update_k_float.argtypes = [
    ndpointer(dtype=ctypes.c_float, ndim=2, flags='C_CONTIGUOUS'), # K
    ndpointer(dtype=ctypes.c_float, ndim=2, flags='C_CONTIGUOUS'), # K_
    ndpointer(dtype=ctypes.c_float, ndim=2, flags='C_CONTIGUOUS'), # C
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # u
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # v
    ctypes.c_float, # epsilon
    ctypes.c_int, # m
    ctypes.c_int  # n
]

lib.update_k_float.restype = None

lib.update_k_double.argtypes = [
    ndpointer(dtype=ctypes.c_double, ndim=2, flags='C_CONTIGUOUS'), # K
    ndpointer(dtype=ctypes.c_double, ndim=2, flags='C_CONTIGUOUS'), # K_
    ndpointer(dtype=ctypes.c_double, ndim=2, flags='C_CONTIGUOUS'), # C
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # u
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # v
    ctypes.c_double, # epsilon
    ctypes.c_int, # m
    ctypes.c_int  # n
]

lib.update_k_double.restype = None

lib.update_R_float.argtypes = [
    ndpointer(dtype=ctypes.c_float, ndim=2, flags='C_CONTIGUOUS'), # R
    ndpointer(dtype=ctypes.c_float, ndim=2, flags='C_CONTIGUOUS'), # K
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # a
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # b
    ctypes.c_int, # m
    ctypes.c_int  # n
]

lib.update_R_float.restype = None

lib.update_R_double.argtypes = [
    ndpointer(dtype=ctypes.c_double, ndim=2, flags='C_CONTIGUOUS'), # R
    ndpointer(dtype=ctypes.c_double, ndim=2, flags='C_CONTIGUOUS'), # K
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # a
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # b
    ctypes.c_int, # m
    ctypes.c_int  # n
]

lib.update_R_double.restype = None

'''
lib.update_a_b_float.argtypes = [
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # a
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # b
    ndpointer(dtype=ctypes.c_float, ndim=2, flags='C_CONTIGUOUS'), # K
    ndpointer(dtype=ctypes.c_float, ndim=2, flags='C_CONTIGUOUS'), # Kt
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # dx
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # dy
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # p
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # q
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # u
    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'), # v
    ctypes.c_float, # lambda1
    ctypes.c_float, # lambda2
    ctypes.c_float, # alpha1
    ctypes.c_float, # alpha2
    ctypes.c_float, # epsilon
    ctypes.c_int, # m
    ctypes.c_int  # n
]

lib.update_a_b_float.restype = None

lib.update_a_b_double.argtypes = [
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # a
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # b
    ndpointer(dtype=ctypes.c_double, ndim=2, flags='C_CONTIGUOUS'), # K
    ndpointer(dtype=ctypes.c_double, ndim=2, flags='C_CONTIGUOUS'), # Kt
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # dx
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # dy
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # p
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # q
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # u
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # v
    ctypes.c_double, # lambda1
    ctypes.c_double, # lambda2
    ctypes.c_double, # alpha1
    ctypes.c_double, # alpha2
    ctypes.c_double, # epsilon
    ctypes.c_int, # m
    ctypes.c_int  # n
]

lib.update_a_b_double.restype = None
'''

lib.step1_process_double.argtypes = [
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # a
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # b
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # old_a
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # old_b
    ndpointer(dtype=ctypes.c_double, ndim=2, flags='C_CONTIGUOUS'), # K
    ndpointer(dtype=ctypes.c_double, ndim=2, flags='C_CONTIGUOUS'), # C
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # dx
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # dy
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # p
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # q
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # u
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # v
    ctypes.c_int, # cur_iter
    ctypes.c_int, # max_iter
    ctypes.c_int, # iters
    ctypes.c_double, # tau
    ctypes.c_double, # lambda1
    ctypes.c_double, # lambda2
    ctypes.c_double, # alpha1
    ctypes.c_double, # alpha2
    ctypes.c_double, # epsilon
    ctypes.c_int, # m
    ctypes.c_int  # n
]

lib.step1_process_double.restype = ctypes.c_int

lib.update_process_double.argtypes = [
    ndpointer(dtype=ctypes.c_double, ndim=2, flags='C_CONTIGUOUS'), # R
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # a
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # b
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # old_a
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # old_b
    ndpointer(dtype=ctypes.c_double, ndim=2, flags='C_CONTIGUOUS'), # K
    ndpointer(dtype=ctypes.c_double, ndim=2, flags='C_CONTIGUOUS'), # _K
    ndpointer(dtype=ctypes.c_double, ndim=2, flags='C_CONTIGUOUS'), # C
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # dx
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # dy
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # p
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # q
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # u
    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'), # v
    ctypes.c_int, # epsilon_scalings
    ctypes.c_int, # cur_epsilon_scaling
    ctypes.c_int, # batch_size
    ctypes.c_double, # epsilon
    ctypes.c_double, # threshold
    ctypes.c_double, # tau
    ctypes.c_double, # lambda1
    ctypes.c_double, # lambda2
    ctypes.c_double, # alpha1
    ctypes.c_double, # alpha2
    ctypes.c_int, # m
    ctypes.c_int  # n
]

lib.update_process_double.restype = ctypes.c_double

def dummy_c(C, K, R, dx, dy, p, q, a, b, epsilon, lambda1, lambda2, use_float=False):
    m, n = C.shape

    if use_float:
        return lib.dummy_float(
            np.ascontiguousarray(C, dtype=ctypes.c_float),
            np.ascontiguousarray(K, dtype=ctypes.c_float),
            np.ascontiguousarray(R, dtype=ctypes.c_float),
            np.ascontiguousarray(dx, dtype=ctypes.c_float),
            np.ascontiguousarray(dy, dtype=ctypes.c_float),
            np.ascontiguousarray(p, dtype=ctypes.c_float),
            np.ascontiguousarray(q, dtype=ctypes.c_float),
            np.ascontiguousarray(a, dtype=ctypes.c_float),
            np.ascontiguousarray(b, dtype=ctypes.c_float),
            epsilon, lambda1, lambda2, m, n)
    else:
        return lib.dummy_double(
            np.ascontiguousarray(C, dtype=ctypes.c_double),
            np.ascontiguousarray(K, dtype=ctypes.c_double),
            np.ascontiguousarray(R, dtype=ctypes.c_double),
            np.ascontiguousarray(dx, dtype=ctypes.c_double),
            np.ascontiguousarray(dy, dtype=ctypes.c_double),
            np.ascontiguousarray(p, dtype=ctypes.c_double),
            np.ascontiguousarray(q, dtype=ctypes.c_double),
            np.ascontiguousarray(a, dtype=ctypes.c_double),
            np.ascontiguousarray(b, dtype=ctypes.c_double),
            epsilon, lambda1, lambda2, m, n)

def primal_c(C, K, R, dx, dy, p, q, a, b, epsilon, lambda1, lambda2, use_float=False):
    m, n = C.shape

    if use_float:
        return lib.primal_float(
            np.ascontiguousarray(C, dtype=ctypes.c_float),
            np.ascontiguousarray(K, dtype=ctypes.c_float),
            np.ascontiguousarray(R, dtype=ctypes.c_float),
            np.ascontiguousarray(dx, dtype=ctypes.c_float),
            np.ascontiguousarray(dy, dtype=ctypes.c_float),
            np.ascontiguousarray(p, dtype=ctypes.c_float),
            np.ascontiguousarray(q, dtype=ctypes.c_float),
            np.ascontiguousarray(a, dtype=ctypes.c_float),
            np.ascontiguousarray(b, dtype=ctypes.c_float),
            epsilon, lambda1, lambda2, m, n)
    else:
        return lib.primal_double(
            np.ascontiguousarray(C, dtype=ctypes.c_double),
            np.ascontiguousarray(K, dtype=ctypes.c_double),
            np.ascontiguousarray(R, dtype=ctypes.c_double),
            np.ascontiguousarray(dx, dtype=ctypes.c_double),
            np.ascontiguousarray(dy, dtype=ctypes.c_double),
            np.ascontiguousarray(p, dtype=ctypes.c_double),
            np.ascontiguousarray(q, dtype=ctypes.c_double),
            np.ascontiguousarray(a, dtype=ctypes.c_double),
            np.ascontiguousarray(b, dtype=ctypes.c_double),
            epsilon, lambda1, lambda2, m, n)

def dual_c(C, K, R, dx, dy, p, q, a, b, epsilon, lambda1, lambda2, use_float=False):
    m, n = C.shape

    if use_float:
        return lib.dual_float(
            np.ascontiguousarray(C, dtype=ctypes.c_float),
            np.ascontiguousarray(K, dtype=ctypes.c_float),
            np.ascontiguousarray(R, dtype=ctypes.c_float),
            np.ascontiguousarray(dx, dtype=ctypes.c_float),
            np.ascontiguousarray(dy, dtype=ctypes.c_float),
            np.ascontiguousarray(p, dtype=ctypes.c_float),
            np.ascontiguousarray(q, dtype=ctypes.c_float),
            np.ascontiguousarray(a, dtype=ctypes.c_float),
            np.ascontiguousarray(b, dtype=ctypes.c_float),
            epsilon, lambda1, lambda2, m, n)
    else:
        return lib.dual_double(
            np.ascontiguousarray(C, dtype=ctypes.c_double),
            np.ascontiguousarray(K, dtype=ctypes.c_double),
            np.ascontiguousarray(R, dtype=ctypes.c_double),
            np.ascontiguousarray(dx, dtype=ctypes.c_double),
            np.ascontiguousarray(dy, dtype=ctypes.c_double),
            np.ascontiguousarray(p, dtype=ctypes.c_double),
            np.ascontiguousarray(q, dtype=ctypes.c_double),
            np.ascontiguousarray(a, dtype=ctypes.c_double),
            np.ascontiguousarray(b, dtype=ctypes.c_double),
            epsilon, lambda1, lambda2, m, n)

def compute_duality_gap_c(C, K, R, dx, dy, p, q, a, b, epsilon, lambda1, lambda2, use_float=False):
    m, n = C.shape

    if use_float:
        return lib.compute_duality_gap_float(
            np.ascontiguousarray(C, dtype=ctypes.c_float),
            np.ascontiguousarray(K, dtype=ctypes.c_float),
            np.ascontiguousarray(R, dtype=ctypes.c_float),
            np.ascontiguousarray(dx, dtype=ctypes.c_float),
            np.ascontiguousarray(dy, dtype=ctypes.c_float),
            np.ascontiguousarray(p, dtype=ctypes.c_float),
            np.ascontiguousarray(q, dtype=ctypes.c_float),
            np.ascontiguousarray(a, dtype=ctypes.c_float),
            np.ascontiguousarray(b, dtype=ctypes.c_float),
            epsilon, lambda1, lambda2, m, n)
    else:
        return lib.compute_duality_gap_double(
            C, #np.ascontiguousarray(C, dtype=ctypes.c_double),
            K, #np.ascontiguousarray(K, dtype=ctypes.c_double),
            R, #np.ascontiguousarray(R, dtype=ctypes.c_double),
            dx, #np.ascontiguousarray(dx, dtype=ctypes.c_double),
            dy, #np.ascontiguousarray(dy, dtype=ctypes.c_double),
            p, #np.ascontiguousarray(p, dtype=ctypes.c_double),
            q, #np.ascontiguousarray(q, dtype=ctypes.c_double),
            a, #np.ascontiguousarray(a, dtype=ctypes.c_double),
            b, #np.ascontiguousarray(b, dtype=ctypes.c_double),
            epsilon, lambda1, lambda2, m, n)

def update_K_c(K, _K, C, u, v, epsilon, use_float=False):
    m, n = C.shape

    if use_float:
        lib.update_k_float(
            K,
            _K,
            C,
            u,
            v,
            epsilon,
            m,
            n
        )

    else:
        lib.update_k_double(
            K,
            _K,
            C,
            u,
            v,
            epsilon,
            m,
            n
        )

def update_R_c(R, K, a, b, use_float=False):
    m, n = K.shape

    if use_float:
        lib.update_R_float(
            R,
            K,
            a,
            b,
            m,
            n
        )

    else:
        lib.update_R_double(
            R,
            K,
            a,
            b,
            m,
            n
        )

def update_a_b_c(a, b, K, Kt, dx, dy, p, q, u, v, lambda1, lambda2, alpha1, alpha2, epsilon, use_float=False):
    m, n = K.shape

    if use_float:
        lib.update_a_b_float(
            a,
            b,
            K,
            Kt,
            dx,
            dy,
            p,
            q,
            u,
            v,
            lambda1,
            lambda2,
            alpha1,
            alpha2,
            epsilon,
            m, n
        )
    else:
        lib.update_a_b_double(
            a,
            b,
            K,
            Kt,
            dx,
            dy,
            p,
            q,
            u,
            v,
            lambda1,
            lambda2,
            alpha1,
            alpha2,
            epsilon,
            m, n
        )

def step1_process_c(
    a, b, old_a, old_b, K, C, dx, dy, p, q, u, v,
    cur_iter, max_iter, iters, tau, lambda1, lambda2, alpha1, alpha2, epsilon):
    m, n = K.shape

    return lib.step1_process_double(
            a,
            b,
            old_a,
            old_b,
            K,
            C,
            dx,
            dy,
            p,
            q,
            u,
            v,
            cur_iter,
            int(max_iter),
            iters,
            float(tau),
            lambda1,
            lambda2,
            alpha1,
            alpha2,
            epsilon,
            m,
            n
        )

def update_process_c(
    R, a, b, old_a, old_b, K, _K, C,
    dx, dy, p, q, u, v,
    epsilon_scaling, cur_epsilon_scaling, batch_size, epsilon, threshold,
    tau, lambda1, lambda2, alpha1, alpha2,
    cur_iter, max_iter
    ):
    m, n = K.shape

    return lib.update_process_double(
        R, a, b, old_a, old_b, K, _K, C,
        dx, dy, p, q, u, v,
        epsilon_scaling, cur_epsilon_scaling, batch_size, epsilon, threshold,
        float(tau), lambda1, lambda2, alpha1, alpha2,
        cur_iter, int(max_iter), m, n
    )



import os
from time import time
import random
import numpy as np
import pandas as pd
import scanpy as sc
import anndata
from sklearn.preprocessing import StandardScaler
from sklearn.cluster import KMeans

import torch
from torch import optim
from torch.utils.data import Dataset, TensorDataset
from torch_geometric.data import Data
from torch_geometric.utils import dense_to_sparse
from torch_geometric.loader import NeighborLoader
from tqdm.auto import tqdm
from collections import OrderedDict

# --- load my package
from SpaDOT.utils.OT_loss.ot_solvers import compute_transport_map # Waddington, faster C version
from SpaDOT.utils import _utils
from SpaDOT.model import SpaDOT

class MyDataset(Dataset):
    def __init__(self, pos, counts, tp_ix, dtype=torch.float64):
        self.dataset = TensorDataset(torch.tensor(pos, dtype=dtype), 
                                     torch.tensor(counts, dtype=dtype), 
                                     torch.tensor(tp_ix, dtype=torch.int))
    def __getitem__(self, index):
        pos, counts, tp_ix  = self.dataset[index]
        return pos, counts, tp_ix
    def __len__(self):
        return len(self.dataset)
    

def prepare_dataloader(adata, model_config):
    # --- obtain location and time point information
    loc = _obtain_tp_loc_info(adata)
    # sample some inducing points
    inducing_idx = random.sample(range(loc.shape[0]), model_config['inducing_point_nums']) 
    inducing_points = loc[inducing_idx, :]
    # --- generate mapper between tp_mat index and real timepoint -> in case the timepoint is not numeric or continous
    idx_to_tp = dict(enumerate(model_config['timepoints']))
    tp_to_idx = {v:k for k,v in idx_to_tp.items()} 
    idx = np.argmax(loc[:, 2:], axis=1)    
    tp_list = [idx_to_tp[_] for _ in idx]
    tp_index_dict = dict() # tp index in timepoint list
    for tp in model_config['timepoints']:
        tp_ix = [_ for _ in range(len(tp_list)) if tp_list[_] == tp]
        tp_index_dict[tp] = tp_ix
    # create inducing point dict / number of training sample dict for separate decoder
    inducing_points_dict, N_train_dict = OrderedDict(), OrderedDict()
    inducing_points_tp = np.argmax(inducing_points[:, 2:], axis=1)
    for _, tp in enumerate(model_config['timepoints']):
        idx = tp_to_idx[tp]
        tp_inducing_idx = np.where(inducing_points_tp == idx)[0]
        inducing_points_dict[tp] = inducing_points[tp_inducing_idx, :2]
        N_train_dict[tp] = np.sum(adata.obs['timepoint'] == tp) # number of training data in each timepoint
    # --- generate timepoint-specific dataloader
    dataloaders_dict, adj_dict, dataset_dict = OrderedDict(), OrderedDict(), OrderedDict()
    for tp in model_config['timepoints']:
        tp_ix = tp_index_dict[tp]
        tp_adata = adata[tp_ix].copy()
        tp_loc = loc[tp_ix, :2]
        tp_dataset = MyDataset(tp_loc, tp_adata.X, tp_ix)
        dataset_dict[tp] = tp_dataset
        # --- Neighborloader for graph data
        _utils._Cal_Spatial_Net(tp_adata, k_cutoff=min(model_config['max_neighbors'], model_config['knn_cutoff']*round(1/1000*tp_adata.n_obs)),
                               max_neigh=model_config['max_neighbors'])
        tp_adj = torch.tensor(tp_adata.uns['adj'], dtype=model_config['dtype'])
        tp_edge_index, _ = dense_to_sparse(tp_adj)
        tp_graph_data = Data(
            x=torch.tensor(tp_adata.X, dtype=model_config['dtype']),
            # edge_index=tp_edge_index.clone().to(dtype=torch.long), # torch copy warning
            edge_index=torch.tensor(tp_edge_index, dtype=torch.long),
            data_index=torch.tensor(tp_ix, dtype=torch.int),
            loc=torch.tensor(tp_loc, dtype=model_config['dtype'])
        )
        tp_graph_loader = NeighborLoader(
            tp_graph_data, num_neighbors=[max(30, model_config['knn_cutoff']*round(1/1000*tp_adata.n_obs))] * 2, 
            batch_size = model_config['batch_size'],
            subgraph_type="induced",
            worker_init_fn=_utils.seed_worker
        )
        dataloaders_dict[tp] = tp_graph_loader 
        adj_dict[tp] = tp_adj
    return {
        "inducing_points": inducing_points_dict,
        "N_train": N_train_dict,
        "dataloaders": dataloaders_dict,
        "adjacency_matrices": adj_dict,
        "datasets": dataset_dict
    }


# --- obtain latent representations
def get_latent(model, model_config, adata, dataloader_dict):
    '''Obtain latent representations
    :param model: SpaDOT model
    :param model_config: model configuration
    :return: latent representations
    '''
    model.eval()
    with torch.no_grad():
        latent_adata_list = list()
        for tp in model_config['timepoints']:
            tp_loc, tp_y, tp_idx = dataloader_dict['datasets'][tp].dataset.tensors
            tp_adj = dataloader_dict['adjacency_matrices'][tp]
            tp_edge_index, _ = dense_to_sparse(tp_adj)
            latent_samples = model.all_latent_samples(tp_loc, tp_y, tp_edge_index, tp)  
            tp_latent_adata = sc.AnnData(latent_samples, obs=adata[tp_idx.numpy()].obs)
            tp_latent_adata.obsm['spatial'] = adata.obsm['spatial'][tp_idx.numpy()]
            latent_adata_list.append(tp_latent_adata)
        latent_adata = anndata.concat(latent_adata_list)
    return latent_adata

def _obtain_tp_loc_info(adata):
    '''
    Obtain the location and time point information from the adata object.
    The location is standardized and concatenated with the time point information.
    The time point information is one-hot encoded.
    '''
    # in case the timepoint is not numeric
    adata.obs['timepoint_numeric'] = adata.obs['timepoint'].astype('category').cat.codes
    tp_info = np.array(adata.obs['timepoint_numeric']).astype('int')
    tp_mat = np.zeros((tp_info.size, tp_info.max()+1))
    tp_mat[np.arange(tp_info.size), tp_info] = 1
    n_tp = tp_mat.shape[1]
    # scale locations per time point
    loc = adata.obsm['spatial']
    loc_scaled = np.zeros(loc.shape, dtype=np.float64)
    for i in range(n_tp):
        scaler = StandardScaler()
        tp_loc = loc[tp_mat[:,i]==1, :]
        tp_loc = scaler.fit_transform(tp_loc)
        loc_scaled[tp_mat[:,i]==1, :] = tp_loc
    loc = loc_scaled
    loc = np.concatenate((loc, tp_mat), axis=1)
    return loc


def _beta_cycle_linear(n_iter, start=0.0, stop=1, n_cycle=10, ratio=1):
    L = np.ones(n_iter) * stop
    period = n_iter/n_cycle
    step = (stop-start)/(period*ratio) # linear schedule
    for c in range(n_cycle):
        v, i = start, 0
        while v <= stop and (int(i+c*period) < n_iter):
            L[int(i+c*period)] = v
            v += step
            i += 1
    return L 

def train_SpaDOT(dataloader_dict, model_config):
    SpaDOT_model = SpaDOT.SpaDOT(model_config, dataloader_dict).to(model_config['device'])
    print(SpaDOT_model)
    optimizer = optim.AdamW(filter(lambda p: p.requires_grad, SpaDOT_model.parameters()), lr=model_config['lr'])

    # --- prepare training
    beta1s = _beta_cycle_linear(model_config['maxiter'], stop=model_config['beta1'])
    tp_indexed_list = list(enumerate(model_config['timepoints']))
    
    # --- initialize dictionary to store loss
    loss_dict = OrderedDict()
    loss_names = ['elbo', 'Recon', 'SVGP_KL', 'GAT_KL', 'alignment', 'KMeans', 'OT']
    for epoch in range(model_config['maxiter']):
        loss_dict[epoch] = OrderedDict()
        for name in loss_names:
            loss_dict[epoch][name] = 0

    print("Training SpaDOT model...")
    train_starttime = time()
    for epoch in tqdm(range(model_config['maxiter'])):
        beta1 = beta1s[epoch]
        # --- start training
        SpaDOT_model.train()
        ep_starttime = time()
        tp_loss_dict = OrderedDict()
        # --- random shuffle timepoints
        random.shuffle(tp_indexed_list) 
        for tp_i, tp in tp_indexed_list:
            tp_loss_dict[tp] = OrderedDict()
            for name in loss_names:
                tp_loss_dict[tp][name] = 0
            tp_dataloader = dataloader_dict['dataloaders'][tp]
            for _, batch in enumerate(tp_dataloader):
                # y: gene expression; x: location; tp_ix: timepoint index; edge_index: graph edge index
                y_batch, x_batch, tp_ix, edge_index_batch = batch.x, batch.loc, batch.data_index, batch.edge_index
                x_batch, y_batch, edge_index_batch = x_batch.to(model_config['device']), y_batch.to(model_config['device']), edge_index_batch.to(model_config['device'])
                tp_ix, adj_idx,  = tp_ix[:batch.batch_size], batch.n_id[:batch.batch_size] # subset to seed nodes only
                # --- forward SVGPVAE and GATVAE to obtain latent space
                tp_Recon_val, tp_SVGP_KL_val, tp_GAT_KL_val, tp_alignment_val, tp_p_m = \
                SpaDOT_model.forward(x=x_batch, y=y_batch, edge_index=edge_index_batch, 
                                     tp=tp, batch_size=batch.batch_size)
                # --- calculate KMeans loss from latent space
                tp_KMeans_val = torch.tensor(0, dtype=model_config['dtype'], device=model_config['device'])
                if epoch >= 1: # add Kmeans loss after the first epoch once the latent space is generated
                    kmeans_loss = _compute_kmeans_loss(SpaDOT_model, model_config, tp, tp_ix, tp_p_m)
                    tp_KMeans_val += kmeans_loss
                # --- calculate OT loss from latent space
                tp_OT_val = torch.tensor(0, dtype=model_config['dtype'], device=model_config['device'])
                if epoch >= model_config['ot_epoch']:
                    if tp_i != 0:
                        tp_loss = _compute_OT_loss(SpaDOT_model, model_config, tp, tp_ix, tp_p_m, model_config['timepoints'][tp_i-1])
                        tp_OT_val += tp_loss
                tp_elbo_val = model_config['lambda1'] * tp_Recon_val - \
                    beta1 * tp_SVGP_KL_val + \
                    model_config['beta2'] * tp_GAT_KL_val + \
                    model_config['omiga1'] * tp_alignment_val + \
                    model_config['omiga2'] * tp_KMeans_val + \
                    model_config['omiga3'] * tp_OT_val
                # --- backward
                optimizer.zero_grad()
                tp_elbo_val.backward(retain_graph=True)
                torch.nn.utils.clip_grad_norm_(SpaDOT_model.parameters(), 0.3) # stabilize training
                optimizer.step()
                # --- update loss in each timepoint
                for name in loss_names:
                    tp_loss_dict[tp][name] += locals()[f'tp_{name}_val'].detach().cpu().item()
            # --- update loss in loss dictionary
            for name in loss_names:
                tp_loss_dict[tp][name] /= len(tp_dataloader)
                loss_dict[epoch][f'{name}'] += tp_loss_dict[tp][name]
        if epoch % 10 == 0:
            print(f'Epoch {epoch + 1}: Training time: {int(time()-ep_starttime)} seconds, ELBO: {loss_dict[epoch]["elbo"]:.8f}, Recon loss: {loss_dict[epoch]["Recon"]:.8f}, SVGP KL loss: {loss_dict[epoch]["SVGP_KL"]:.8f}, GAT KL loss: {loss_dict[epoch]["GAT_KL"]:.8f}, Alignment loss: {loss_dict[epoch]["alignment"]:.8f}, Kmeans loss: {loss_dict[epoch]["KMeans"]:.8f}, OT loss: {loss_dict[epoch]["OT"]:.8f}')
        # --- update Kmeans centroids
        _update_Kmeans(SpaDOT_model, model_config, dataloader_dict)
        # --- update OT matrix 
        if (epoch + 1) % model_config['ot_config']['ot_epochs'] == 0:
            _update_OT_matrix(SpaDOT_model, model_config)

    print('Training finished...')
    print('Training time: %d seconds.' % int(time() - train_starttime))
    loss_df = pd.DataFrame.from_dict(loss_dict)
    return SpaDOT_model, loss_df


# K-Means related
def _compute_kmeans_loss(model, model_config, tp, tp_ix, latent):
    '''Compute KMeans loss between assignment and centroids
    :param tp: time point
    :param tp_ix: time point specific index
    :param latent: GP posterior latent space

    :return: K-Means loss
    '''
    cluster_index_dict = model.kmeans_index_dict
    cluster_assignments = [cluster_index_dict[tp][_.item()] for _ in tp_ix]
    cluster_centers = torch.tensor(model.kmeans_center_dict[tp], dtype=model_config['dtype'], device=model_config['device'])
    # Compute K-means loss (mean of squared distances)
    kmeans_loss = torch.sum(torch.norm(latent - cluster_centers[cluster_assignments]) ** 2 / latent.shape[1] / len(set(cluster_assignments)) )
    return kmeans_loss

def _update_Kmeans(model, model_config, dataloader_dict):
    model.eval()
    with torch.no_grad():
        for tp in model_config['timepoints']:
            tp_loc, tp_y, tp_idx = dataloader_dict['datasets'][tp].dataset.tensors
            tp_adj = dataloader_dict['adjacency_matrices'][tp]
            tp_edge_index, _ = dense_to_sparse(tp_adj)
            latent_samples = model.all_latent_samples(tp_loc, tp_y, tp_edge_index, tp)  
            # adaptive KMeans
            tp_kmeans = KMeans(n_clusters=model_config['n_clusters'], 
                               random_state=model_config['seed'], 
                               n_init=10).fit(latent_samples)
            model.kmeans_center_dict[tp] = tp_kmeans.cluster_centers_
            model.kmeans_cluster_dict[tp] = tp_kmeans.labels_.tolist()
            model.kmeans_index_dict[tp] = dict(zip(tp_idx.numpy(), tp_kmeans.labels_))

# OT related
def _compute_OT_loss(model, model_config, cur_tp, tp_ix, tp_p_m, prev_tp):
    '''Compute OT loss between timepoints
    :param model:  model
    :param tp: current timepoints
    :param tp_ix: index of current tp to obtain hidden clusters
    :param tp_p_m: posterior mean of current timepoints
    :prev_tp: previous timepoint
    :return: OT transport cost
    '''
    cur_tp_cluster_latent = dict()
    cur_tp_clusters = [model.kmeans_index_dict[cur_tp][_] for _ in tp_ix.numpy()]
    cluster_list = list(set(model.kmeans_cluster_dict[cur_tp]))
    cluster_list.sort()
    for idx, cluster in enumerate(cur_tp_clusters):
        if cluster in cur_tp_cluster_latent:
            cur_tp_cluster_latent[cluster].append(tp_p_m[idx])
        else:
            cur_tp_cluster_latent[cluster] = [tp_p_m[idx]]
    cur_tp_cluster_center = []
    for cluster in cluster_list:
        if cluster not in cur_tp_cluster_latent:
            cur_tp_cluster_center.append(torch.tensor(model.kmeans_center_dict[cur_tp][cluster], 
                                                      dtype=model_config['dtype'], device=model_config['device']))
        else:
            cur_tp_cluster_center.append(torch.mean(torch.stack(cur_tp_cluster_latent[cluster]), dim=0))
    cur_tp_cluster_center = torch.stack(cur_tp_cluster_center)
    gamma = model.gammas[f"{prev_tp}_{cur_tp}"] # obtain transport plan
    gamma = gamma / gamma.sum(axis=1, keepdims=True) # normalize rows
    gamma = np.nan_to_num(gamma, nan=0.0, posinf=0.0, neginf=0.0)  # prune invalid values
    gamma = torch.tensor(gamma, dtype=model_config['dtype']).to(model_config['device'])
    cost_matrix = torch.cdist(
        torch.tensor(model.kmeans_center_dict[prev_tp], dtype=model_config['dtype']).to(model_config['device']), 
        cur_tp_cluster_center, 
        p=2) # (prev_tp_cluster, cur_tp_cluster)
    transport_cost = torch.mean(gamma*cost_matrix)
    return transport_cost

def _update_OT_matrix(model, model_config):
    '''Compute OT matrix
    '''
    model.eval()
    timepoints = model_config['timepoints']
    for tp_i, tp in enumerate(timepoints):
        if tp_i == len(timepoints) - 1: break
        cur_tp = tp
        next_tp = timepoints[tp_i+1]
        gamma = compute_transport_map(model.kmeans_center_dict[cur_tp],
                                       model.kmeans_center_dict[next_tp], 
                                       model_config['ot_config'], G=None)
        model.gammas[f"{timepoints[tp_i]}_{timepoints[tp_i+1]}"] = gamma